--------------------------------------------------------------------------------
-- SubModule GPSReceiver
-- Created   12/19/2007 6:58:33 PM
--------------------------------------------------------------------------------
library IEEE;
use IEEE.Std_Logic_1164.all;
use IEEE.Numeric_Std.all;

entity GPSReceiver is port
   (
     CLK        : in    std_logic;
     RST        : in    std_logic;
     ONEPPS     : in    std_logic;
     PRESENT    : in    std_logic;
     INIT       : out   std_logic;
     HALT       : out   std_logic;

     RDY        : in    std_logic;
     ERR        : in    std_logic;
     D          : in    std_logic_vector(7 downto 0);

     LOAD       : out   std_logic;
     FIRST      : out   std_logic;
     DATE       : out   std_logic_vector(7 downto 0);
     OK         : out   std_logic;
     CRCERR     : out   std_logic;
     LEAP       : out   std_logic_vector(15 downto 0);
     STAT       : out   std_logic_vector(255 downto 0);
     LOCK       : out   std_logic;
     MESG1      : out   std_logic_vector(7 downto 0);
     MESG2      : out   std_logic_vector(7 downto 0);
     MESG3      : out   std_logic_vector(7 downto 0);
     MESG4      : out   std_logic_vector(7 downto 0)
   );
end GPSReceiver;
--------------------------------------------------------------------------------

--------------------------------------------------------------------------------
architecture Structure of GPSReceiver is

-- Component Declarations

-- Constant Declarations
   constant ret1 : string := "Hb";
   constant ret2 : string := "Ha";
   constant ret3 : string := "Bj";
   constant ret4 : string := "Gd";

-- Type Declaration
   type FSM_state is (IDLE, HEADER, MSG1, MSG2, MSG3, MSG4, TAIL);

-- Signal Declarations
   signal oldonepps               : std_logic;
   signal timeout                 : unsigned (27 downto 0); -- 4 seconds
   signal expired                 : std_logic;
   signal state                   : FSM_state;
   signal crc                     : std_logic_vector (7 downto 0);
   signal dold                    : std_logic_vector (7 downto 0);
   signal count                   : unsigned (7 downto 0);
   signal year                    : unsigned (15 downto 0);
   signal info                    : std_logic_vector(255 downto 0);
   signal leapinfo                : std_logic_vector (7 downto 0);
   signal leappen                 : std_logic_vector (1 downto 0);
   signal posmode                 : std_logic_vector (1 downto 0);
   signal status_locked           : std_logic;
   signal msg1cnt                 : unsigned (7 downto 0);
   signal msg2cnt                 : unsigned (7 downto 0);
   signal msg3cnt                 : unsigned (7 downto 0);
   signal msg4cnt                 : unsigned (7 downto 0);

-- Functions
   function check_string (s : in string; v : in std_logic_vector (15 downto 0))
                          return boolean is
      variable i : integer;
      --variable c : character;
      variable w : std_logic_vector (7 downto 0);
      variable ret : boolean;
   begin
      ret := true;
      for i in 1 to 2 loop
         --c := s (i);
         w := std_logic_vector (To_unsigned (character'pos(s(i)), 8));
         if w /= v (i*8-1 downto (i-1)*8) then
            ret := false;
         end if;
      end loop;
      return ret;
   end function check_string;

begin
   MESG1 <= std_logic_vector (msg1cnt);
   MESG2 <= std_logic_vector (msg2cnt);
   MESG3 <= std_logic_vector (msg3cnt);
   MESG4 <= std_logic_vector (msg4cnt);

   -- check if we received a message in the past 4 seconds
   oldonepps <= ONEPPS when rising_edge (CLK);
   process (RST, PRESENT, CLK) is
   begin
      if (RST = '1') or (PRESENT = '0') then
         expired <= '0';
         timeout <= (others => '1');
      elsif rising_edge (CLK) then
         -- check for MSG2 and correct UTC time mode
         if (state = MSG2) and (leapinfo(7) = '0') then
            timeout <= (others => '1');
            expired <= '0';
         else
            if timeout = To_unsigned (0, 28) then
                expired <= '1';
            else
                expired <= '0';
            end if;
            timeout <= timeout - 1;
         end if;
      end if;
   end process;

   -- receive the message
   process (RST, CLK) is
   begin
      if (RST = '1') then
         state <= IDLE;
         LOAD <= '0';
         FIRST <= '0';
         HALT <= '0';
         DATE <= (others => '0');
         OK <= '0';
         CRCERR <= '0';
         crc <= (others => '0');
         leapinfo <= (others => '0');
         leappen <= (others => '0');
         posmode <= (others => '0');
         info <= (others => '0');
         -- LOCK <= '0';
         status_locked <= '0';
         msg1cnt <= (others => '0');
         msg2cnt <= (others => '0');
         msg3cnt <= (others => '0');
         msg4cnt <= (others => '0');
      elsif rising_edge (CLK) then
         LOAD <= '0';
         FIRST <= '0';
         HALT <= '0';
         OK <= '0';
         CRCERR <= '0';
         case state is
            when IDLE =>
               if RDY = '1' then
                  dold <= d;
                  if check_string ("@@", d & dold) then
                     state <= HEADER;
                     count <= X"01";
                     crc <= (others => '0');
                  end if;
               elsif expired = '1' then
                  status_locked <= '0';
               end if;

            -- message header
            when HEADER =>
               if RDY = '1' then
                  dold <= d;
                  crc <= crc xor d;
                  if count = X"00" then
                     -- Hb message
                     if check_string (ret1, d & dold) then
                        state <= IDLE;
                        HALT <= '1';
                        msg1cnt <= msg1cnt + 1;
 --                       state <= MSG1;
 --                       count <= To_unsigned (47, 8);
                     -- Ha message
                     elsif check_string (ret2, d & dold) then
                        state <= MSG2;
                        count <= To_unsigned (147, 8);
                        msg2cnt <= msg2cnt + 1;
                     -- Bj message
                     elsif check_string (ret3, d & dold) then
                        state <= MSG3;
                        count <= To_unsigned (1, 8);
                        msg3cnt <= msg3cnt + 1;
                     elsif check_string (ret4, d & dold) then
                        state <= MSG4;
                        count <= To_unsigned (1, 8);
                        msg4cnt <= msg4cnt + 1;
                     else
                        state <= IDLE;
                     end if;
                  else
                     count <= count - 1;
                  end if;
               end if;

            -- message 1: Hb
            when MSG1 =>
               if RDY = '1' then
                  dold <= d;
                  crc <= crc xor d;
                  if count /= X"00" then
                     count <= count - 1;
                  end if;
                  if count = To_unsigned (47, 8) then
                     -- month
                     DATE <= d;
                     FIRST <= '1';
                     LOAD <= '1';
                  elsif count = To_unsigned (46, 8) then
                     -- day
                     DATE <= d;
                     LOAD <= '1';
                  elsif count = To_unsigned (44, 8) then
                     -- year
                     DATE <= std_logic_vector (year (7 downto 0));
                     if year (15 downto 8) = X"00" then
                        LOAD <= '1';
                     else
                        state <= IDLE;
                     end if;
                  elsif count = To_unsigned (43, 8) then
                     -- hour
                     DATE <= d;
                     LOAD <= '1';
                  elsif count = To_unsigned (42, 8) then
                     -- minute
                     DATE <= d;
                     LOAD <= '1';
                  elsif count = To_unsigned (41, 8) then
                     -- second
                     DATE <= d;
                     LOAD <= '1';
                  elsif (count <= To_unsigned (36, 8)) and
                        (count >= To_unsigned (25, 8)) then
                     -- position
                     info <= info (247 downto 0) & d;
                  elsif (count <= To_unsigned (20, 8)) and
                        (count >= To_unsigned ( 1, 8)) then
                     -- speed, heading, geometry, satellites, status, tag
                     info <= info (247 downto 0) & d;
                  elsif count = X"00" then
                     state <= TAIL;
                  end if;
               end if;

            -- message 2: Ha
            when MSG2 =>
               if RDY = '1' then
                  dold <= d;
                  crc <= crc xor d;
                  if count /= X"00" then
                     count <= count - 1;
                  end if;
                  if count = To_unsigned (147, 8) then
                     -- month
                     DATE <= d;
                     FIRST <= '1';
                     LOAD <= '1';
                  elsif count = To_unsigned (146, 8) then
                     -- day
                     DATE <= d;
                     LOAD <= '1';
                  elsif count = To_unsigned (144, 8) then
                     -- year
                     DATE <= std_logic_vector (year (7 downto 0));
                     if year (15 downto 8) = X"00" then
                        LOAD <= '1';
                     else
                        state <= IDLE;
                     end if;
                  elsif count = To_unsigned (143, 8) then
                     -- hour
                     DATE <= d;
                     LOAD <= '1';
                  elsif count = To_unsigned (142, 8) then
                     -- minute
                     DATE <= d;
                     LOAD <= '1';
                  elsif count = To_unsigned (141, 8) then
                     -- second
                     DATE <= d;
                     LOAD <= '1';
                  elsif (count <= To_unsigned (136, 8)) and
                        (count >= To_unsigned (125, 8)) then
                     -- position
                     info <= info (247 downto 0) & d;
                  elsif (count <= To_unsigned (104, 8)) and
                        (count >= To_unsigned ( 95, 8)) then
                     -- speed, heading, geometry, satellites
                     info <= info (247 downto 0) & d;
                  elsif (count <= To_unsigned ( 22, 8)) and
                        (count >= To_unsigned ( 19, 8)) then
                     -- status
                     info <= info (247 downto 0) & d;
                  elsif count = To_unsigned (10, 8) then
                     -- leap seconds
                     leapinfo <= d;
                  elsif (count <= To_unsigned (  6, 8)) and
                        (count >= To_unsigned (  1, 8)) then
                     -- tag
                     info <= info (247 downto 0) & d;
                  elsif count = X"00" then
                     state <= TAIL;
                  end if;
               end if;

            -- message 3: Bj
            when MSG3 =>
               if RDY = '1' then
                  crc <= crc xor d;
                  if count /= X"00" then
                     count <= count - 1;
                  end if;
                  if count = To_unsigned (1, 8) then
                     leappen <= d(1 downto 0);
                  elsif count = X"00" then
                     state <= IDLE;
                  end if;
               end if;

            -- message 4: Gd
            when MSG4 =>
               if RDY = '1' then
                  crc <= crc xor d;
                  if count /= X"00" then
                     count <= count - 1;
                  end if;
                  if count = To_unsigned (1, 8) then
                     posmode <= d(1 downto 0);
                  elsif count = X"00" then
                     state <= IDLE;
                  end if;
               end if;

            -- check CRC
            when TAIL =>
               if crc = X"00" then
                  OK <= '1';
                  status_locked <= info(79) and (info(78) xnor info(77)) and
                                   not info(67);
               else
                  CRCERR <= '1';
               end if;
               state <= IDLE;
         end case;
      end if;
   end process;
   year <= unsigned (dold & d) - To_unsigned (2000, 16);
   LEAP(15 downto 14) <= leapinfo (7 downto 6);
   LEAP(13 downto 12) <= leappen;
   LEAP(11 downto  6) <= (others => '0');
   LEAP( 5 downto  0) <= leapinfo (5 downto 0);
   STAT <= info;
   LOCK <= status_locked;
   INIT <= expired;

end Structure;
--------------------------------------------------------------------------------
