--------------------------------------------------------------------------------
-- SubModule MasterDiag
-- Created   8/1/2008 9:28:50 AM
--------------------------------------------------------------------------------
library IEEE;
use IEEE.Std_Logic_1164.all;
use IEEE.Numeric_Std.all;

entity MasterDiag is port
   (
     GPS              : in    std_logic_vector(31 downto 0);
     ADDR             : in    std_logic_vector(31 downto 0);
     LEAP             : in    std_logic_vector(15 downto 0);

     GPS1PPS          : in    std_logic_vector(31 downto 0);
     EXT1PPS          : in    std_logic_vector(31 downto 0);
     UP1PPS           : in    std_logic_vector(31 downto 0);

     SW               : in    std_logic_vector(8 downto 1);

     FOUP             : in    std_logic_vector(15 downto 0);
     FOLOS            : in    std_logic_vector(15 downto 0);
     FOMISSING        : in    std_logic_vector(15 downto 0);
     FODLYERR         : in    std_logic_vector(15 downto 0);

     LOS              : in    std_logic;
     UP               : in    std_logic;
     ERRC             : in    std_logic_vector(3 downto 0);
     CRCERR           : in    std_logic_vector(7 downto 0);
     VCXO_CTRL        : in    std_logic_vector(15 downto 0);

     OCXO_PRESENT     : in    std_logic;
     OCXO_LOCKED      : in    std_logic;
     OCXO_ERR         : in    std_logic_vector(31 downto 0);
     OCXO_CTRL        : in    std_logic_vector(15 downto 0);

     EXT1PPS_ON       : in    std_logic;

     GPS_ON           : in    std_logic;
     GPS_LOCKED       : in    std_logic;
     GPS_ERROR        : in    std_logic;
     GPS_ERRC         : in    std_logic_vector(2 downto 0);

     A                : in    std_logic_vector(3 downto 0);
     D                : out   std_logic_vector(31 downto 0);
     STAT             : out   std_logic_vector(1023 downto 0)
   );
end MasterDiag;
--------------------------------------------------------------------------------

--------------------------------------------------------------------------------
architecture Structure of MasterDiag is

-- Component Declarations

-- Signal Declarations
   signal w0           : std_logic_vector (31 downto 0); -- board id/revision
   signal w1           : std_logic_vector (31 downto 0); -- board serial
   signal w2           : std_logic_vector (31 downto 0); -- software id
   signal w3           : std_logic_vector (31 downto 0); -- software revision
   signal w4           : std_logic_vector (31 downto 0); -- module address
   signal w5           : std_logic_vector (31 downto 0); -- GPS time
   signal w6           : std_logic_vector (31 downto 0); -- Status
   signal w7           : std_logic_vector (31 downto 0); -- Configuration
   signal w8           : std_logic_vector (31 downto 0); -- OCXO Controls
   signal w9           : std_logic_vector (31 downto 0); -- OCXO Error
   signal w10          : std_logic_vector (31 downto 0); -- Uplink 1PPS delay
   signal w11          : std_logic_vector (31 downto 0); -- Ext 1PPS delay
   signal w12          : std_logic_vector (31 downto 0); -- GPS 1PPS delay
   signal w13          : std_logic_vector (31 downto 0); -- Fanout Up/LOS
   signal w14          : std_logic_vector (31 downto 0); -- Fanout Miss/DErr
   signal w15          : std_logic_vector (31 downto 0); -- CRC & GPS error
   signal o            : std_logic_vector (31 downto 0);

begin

   -- board id/revision
   w0 <= X"070011B0";
   -- board serial
   w1 <= X"00000000";
   -- software id
   w2 <= X"080534A0";
   -- software revision
   w3 <= std_logic_vector (To_Unsigned (503, 32));

   -- GPS time
   w4 <= GPS;
   -- address
   w5 <= ADDR;

   -- Status
   w6 (31 downto 16) <= VCXO_CTRL;
   w6 (15 downto 8) <= SW;
   w6 (7) <= '0';
   w6 (6) <= '0';
   w6 (5) <= LOS;
   w6 (4 downto 1) <= ERRC;
   w6 (0) <= UP;

   -- Configuration
   w7(31 downto 16) <= (others => '0');
   w7(15) <= OCXO_PRESENT;
   w7(14) <= EXT1PPS_ON;
   w7(13) <= GPS_ON;
   w7(12) <= GPS_LOCKED;
   w7(11) <= OCXO_LOCKED;
   w7(10) <= OCXO_PRESENT and EXT1PPS_ON;
   w7( 9) <= OCXO_PRESENT and GPS_ON and not EXT1PPS_ON;
   w7( 8) <= not OCXO_PRESENT and UP;
   w7( 7) <= UP;
   w7( 6) <= LOS;
   w7(5 downto 2) <= (others => '0');
   w7( 1) <= '1';
   w7( 0) <= OCXO_PRESENT;

   -- OCXO Controls
   w8(31 downto 16) <= (others => '0');
   w8(15 downto  0) <= OCXO_CTRL;
   -- OCXO Error
   w9 <= OCXO_ERR;

   -- Uplink 1PPS delay
   w10 <= UP1PPS;
   -- Ext 1PPS delay
   w11 <= EXT1PPS;
   -- GPS 1PPS delay
   w12 <= GPS1PPS;

   -- Fanout Up/LOS
   w13 <= FOUP & FOLOS;
   -- Fanout Miss/DErr
   w14 <= FOMISSING & FODLYERR;

   -- CRC Error count
   w15 <= LEAP & X"0" & GPS_ERROR & GPS_ERRC & CRCERR;

   -- output mux
   omux: process (A, w0, w1, w2, w3, w4, w5, w6, w7,
                  w8, w9, w10, w11, w12, w13, w14, w15) is
   begin
      case A is
         when X"0" =>
            o <= w0;
         when X"1" =>
            o <= w1;
         when X"2" =>
            o <= w2;
         when X"3" =>
            o <= w3;
         when X"4" =>
            o <= w4;
         when X"5" =>
            o <= w5;
         when X"6" =>
            o <= w6;
         when X"7" =>
            o <= w7;
         when X"8" =>
            o <= w8;
         when X"9" =>
            o <= w9;
         when X"A" =>
            o <= w10;
         when X"B" =>
            o <= w11;
         when X"C" =>
            o <= w12;
         when X"D" =>
            o <= w13;
         when X"E" =>
            o <= w14;
         when X"F" =>
            o <= w15;
         when others =>
            o <= (others => '0');
      end case;
   end process omux;
   D <= o;

   -- uplink status vector
   STAT(1023 downto 512) <= w0 & w1 &  w2 &  w3 &  w4 &  w5 &  w6 &  w15 &
                         w7 & w8 & w9 & w10 & w11 & w12 & w13 & w14;
   STAT(511 downto 0) <= (others => '0');

end architecture Structure;
--------------------------------------------------------------------------------
