--------------------------------------------------------------------------------
-- SubModule RS422Receiver
-- Created   12/2/2020 11:42:53 AM
--------------------------------------------------------------------------------
library IEEE;
use IEEE.Std_Logic_1164.all;
use IEEE.Numeric_Std.all;

entity RS422Receiver is port
   (
     CLK        : in    std_logic;
     RST        : in    std_logic;
     CNT        : in    std_logic_vector(25 downto 0);
     ERR        : in    std_logic;
     RDY        : in    std_logic;
     D          : in    std_logic_vector(7 downto 0);
     POSHOLD    : out   std_logic;
     UNHOLD     : out   std_logic;
     MESG1      : out   std_logic_vector(31 downto 0);
     MESG2      : out   std_logic_vector(31 downto 0)
   );
end RS422Receiver;
--------------------------------------------------------------------------------

--------------------------------------------------------------------------------
architecture Structure of RS422Receiver is

-- Component Declarations

-- Constant Declarations
   constant lead : string := "@@";
   constant ret1 : string := "poshold+" & cr & lf;
   constant ret2 : string := "poshold-" & cr & lf;
   constant term : string := cr & lf;
   constant maxlen : natural := 10;

-- Type Declaration
   type FSM_state is (IDLE, HEADER, MSG1, MSG2);

-- Signal Declarations
   signal oldonepps               : std_logic;
   signal timeout                 : unsigned (27 downto 0); -- 4 seconds
   signal expired                 : std_logic;

   signal state                   : FSM_state;
   signal dold                    : std_logic_vector (maxlen*8-1 downto 0);
   signal count                   : unsigned (7 downto 0);
   signal msg1cnt                 : unsigned (31 downto 0);
   signal msg2cnt                 : unsigned (31 downto 0);

-- Functions
   function check_string (s : in string; v : in std_logic_vector (maxlen*8-1 downto 0))
                          return boolean is
      variable i : integer;
      variable w : std_logic_vector (7 downto 0);
      variable ret : boolean;
   begin
      ret := true;
      for i in 1 to s'length loop
         w := std_logic_vector (To_unsigned (character'pos(s(i)), 8));
         if w /= v ((s'length - i) * 8 + 7 downto (s'length - i)*8) then
            ret := false;
         end if;
      end loop;
      return ret;
   end function check_string;

begin
   -- receive the message
   process (RST, CLK) is
   begin
      if (RST = '1') then
         state <= IDLE;
         POSHOLD <= '0';
         UNHOLD <= '0';
         dold <= (others => '0');
         msg1cnt <= (others => '0');
         msg2cnt <= (others => '0');
      elsif rising_edge (CLK) then
         POSHOLD <= '0';
         UNHOLD <= '0';
         case state is
            when IDLE =>
               if RDY = '1' then
                  dold <= dold ((maxlen-1)*8-1 downto 0) & D;
                  if check_string (lead, dold ((maxlen-1)*8-1 downto 0) & D) then
                     state <= HEADER;
                     count <= to_unsigned (maxlen - 1 , 8);
                  end if;
               end if;

            -- message header
            when HEADER =>
               if RDY = '1' then
                  dold <= dold ((maxlen-1)*8-1 downto 0) & D;
                  if check_string (ret1, dold ((maxlen-1)*8-1 downto 0) & D) then
                     state <= MSG1;
                  elsif check_string (ret2, dold ((maxlen-1)*8-1 downto 0) & D) then
                     state <= MSG2;
                  elsif check_string (term, dold ((maxlen-1)*8-1 downto 0) & D) then
                     state <= IDLE;
                  elsif count = X"00" then
                     state <= IDLE;
                  else
                     count <= count - 1;
                  end if;
               end if;

            -- message 1: turn position hold on
            when MSG1 =>
               msg1cnt <= msg1cnt + 1;
               POSHOLD <= '1';
               state <= IDLE;

            -- message 2: turn position hold off
            when MSG2 =>
               msg2cnt <= msg2cnt + 1;
               UNHOLD <= '1';
               state <= IDLE;
         end case;
      end if;
   end process;

   MESG1 <= std_logic_vector (msg1cnt);
   MESG2 <= std_logic_vector (msg2cnt);

end Structure;
--------------------------------------------------------------------------------
