--------------------------------------------------------------------------------
-- SubModule TimingLinkFanout
--
-- This module generates the status string of a fanout transceiver.
-- It can be daisy chained in order to link all 16 fan-outs together.
-- An final module describing the local status can be added at the
-- end. The read-out starts with the last module in the chain. The
-- data is available thorugh a simple 8 bit interface which is initialized
-- by setting START high and then requesting a byte by applying ENABLE
-- whenever RDY is high. If no more data is available, DONE goes high.
--
-- The interface to the fiber receiver is thorugh a serial interface. This
-- interface has to be initiliazed every second by applying IN_START, before
-- the first bit is received. Applying IN_CE will read a single bit. When
-- the maximum number of bits have been received IN_DONE goes high. If there
-- was a CRC error in the received data, it will be indicated with the
-- IN_CRCERR bit 1 cycle after DONE goes high.
-- The interface to the fiber transmitter is initilaizd with the given
-- GPS time and address, when OUT_START is applied. The first data bit
-- is immediatly available. Whenever OUT_CE is applied, the next bit is
-- shited out. OUT_DONE goes high, when no more data is available.
--
-- Parameters:
--   CLK: clock
--   RST: reset
--   GPS: current GPS time (will be used for bit output)
--   ADDR: module address (will be used for bit output)
--   A, B: extra 32 bitoutputs i case one needs to send extra information
--   LASTLINK: must be true for the last link in the chain, false otherwise
--             (the last link is the one interfacing the UART)
--
--   ENABLE: this enables the transmission of characters
--   START: set high to start the transmission
--   RDY: a new character is ready to be transmitted
--   DOUT: character data
--   DONE: indicates that we are done
--
--   PREV_ENABLE: daisy chain for enable
--   PREV_START: daisy chain for start
--   PREV_RDY: daisy chain for ready (set low on last block)
--   PREV_DIN: daisy chain for data
--   PREV_DONE: daisy chain for done (set high on last block)
--
--   IN_D: serial data from fiber link
--   IN_CE: clock enable for serial input data
--   IN_START: this restarts the stream to the first bit
--             (must come before the first CE)
--   IN_DONE: enough bits have been received
--   IN_CRCERR: indicates a CRC error on the fiber receiver
--              (valid 1 cycle after IN_DONE goes high)
--
--   OUT_D: serial data to fiber link
--   OUT_CE: clock enable to move to the next bit
--   OUT_START: this restarts the stream to the first bit
--              (must come before the first CE)
--   OUT_DONE: no more data is available
--
-- Created   7/23/2008 DS
--------------------------------------------------------------------------------
Library IEEE;
Use IEEE.Std_Logic_1164.all;
Use IEEE.Numeric_Std.all;

entity TimingLinkFanout is port
   (
     CLK           : in    std_logic;
     RST           : in    std_logic;

     GPS           : in    std_logic_vector(31 downto 0);
     ADDR          : in    std_logic_vector(31 downto 0);
     A             : in    std_logic_vector(31 downto 0);
     B             : in    std_logic_vector(31 downto 0);

     ENABLE        : in    std_logic;
     START         : in    std_logic;
     RDY           : out   std_logic;
     DOUT          : out   std_logic_vector(7 downto 0);
     DONE          : out   std_logic;

     PREV_ENABLE   : out   std_logic;
     PREV_START    : out   std_logic;
     PREV_RDY      : in    std_logic;
     PREV_DIN      : in    std_logic_vector(7 downto 0);
     PREV_DONE     : in    std_logic;

     IN_D          : in    std_logic;
     IN_CE         : in    std_logic;
     IN_START      : in    std_logic;
     IN_DONE       : out   std_logic;
     IN_CRCERR     : out   std_logic;
     IN_CRCOK      : out   std_logic;

     OUT_D         : out   std_logic;
     OUT_CE        : in    std_logic;
     OUT_START     : in    std_logic;
     OUT_DONE      : out   std_logic
   );
end TimingLinkFanout;
--------------------------------------------------------------------------------

--------------------------------------------------------------------------------
architecture Structure of TimingLinkFanout is

-- Constant Declarations
   constant OUTPUT_BITS : integer := 128;  -- not including CRC
   constant INPUT_BITS : integer := 1024; -- including the CRC

-- Constant Declarations
   constant CRC32_POLYNOMIAL : std_logic_vector (31 downto 0) := X"04C11DB7";
   constant CRC32_INIT_VALUE : std_logic_vector (31 downto 0) := X"FFFFFFFF";
   constant CRC32_MATCH_VALUE: std_logic_vector (31 downto 0) := X"C704DD7B";

-- Component Declarations
   component ucrc_ser is
      generic (
         POLYNOMIAL : std_logic_vector;
         INIT_VALUE : std_logic_vector;
         SYNC_RESET : integer range 0 to 1 := 0);  -- use synchronous reset
      port (
         clk_i   : in  std_logic;          -- clock
         rst_i   : in  std_logic;          -- init CRC
         clken_i : in  std_logic;          -- clock enable
         data_i  : in  std_logic;          -- data input
         flush_i : in  std_logic;          -- flush crc
         match_o : out std_logic;          -- CRC match flag
         crc_o   : out std_logic_vector(POLYNOMIAL'length - 1 downto 0));  -- CRC output
   end component;

   component bram_16384_1_8
      port (
        clka     : in  std_logic;
        dina     : in  std_logic_vector (0 downto 0);
        addra    : in  std_logic_vector (13 downto 0);
        wea      : in  std_logic_vector (0 downto 0);
        douta    : out std_logic_vector (0 downto 0);
        clkb     : in  std_logic;
        dinb     : in  std_logic_vector (7 downto 0);
        addrb    : in  std_logic_vector (10 downto 0);
        web      : in  std_logic_vector (0 downto 0);
        doutb    : out std_logic_vector (7 downto 0));
   end component;

-- Type Declarations
   type char_state_type  is (IDLE, RESETING, SHIFT_PREV_WAIT, SHIFT_PREV_READ,
                             SHIFT_PREV_WRITE, SHIFT_D_READ, SHIFT_D_WRITE);
   type in_state_type    is (SHIFT_D, SHIFT_WAIT, SHIFT_DONE);
   type out_state_type   is (SHIFT_D, SHIFT_CRC, SHIFT_DONE);

-- Signals
   signal char_addr      : std_logic_vector (10 downto 0);
   signal char_dout      : std_logic_vector (7 downto 0);
   signal char_done      : std_logic;
   signal char_d         : std_logic_vector (7 downto 0);
   signal char_prev_en   : std_logic;
   signal char_rdy       : std_logic;
   signal char_count     : unsigned (15 downto 0);
   signal char_state     : char_state_type;

   signal in_din         : std_logic_vector (0 downto 0);
   signal in_addr        : std_logic_vector (13 downto 0);
   signal in_we          : std_logic_vector (0 downto 0);
   signal in_crcmatch    : std_logic;
   signal in_count       : unsigned (15 downto 0);
   signal in_state       : in_state_type;
   signal in_end         : std_logic;
   signal in_timeout     : unsigned (27 downto 0);
   signal in_to          : std_logic;

   signal out_reg        : std_logic_vector (127 downto 0);
   signal out_data       : std_logic;
   signal out_crc_flush  : std_logic;
   signal out_crc        : std_logic_vector (31 downto 0);
   signal out_count      : unsigned (15 downto 0);
   signal out_state      : out_state_type;

begin

   -----------------------------------------------------------------------------
   -- CHARACTER INTERFACE
   -----------------------------------------------------------------------------

   -- character interface state machine
   char_fsm: process (CLK) is
   begin
      if rising_edge (CLK) then
         if RST = '1' then
            char_state <= IDLE;
            char_done <= '1';
            char_prev_en <= '0';
            char_rdy <= '0';
         elsif START = '1' then
            char_state <= RESETING;
            char_done <= '0';
            char_prev_en <= '0';
            char_rdy <= '0';
         else
            char_prev_en <= '0';
            case char_state is
               when IDLE =>
                  -- nothing
                  char_rdy <= '0';
                  char_done <= '1';
                  char_count <= To_Unsigned (0, 16);

               -- wait for previous module to reset
               when RESETING =>
                  char_done <= '0';
                  char_rdy <= '0';
                  char_count <= To_Unsigned (0, 16);
                  char_state <= SHIFT_PREV_WAIT;

               -- shift out data from previous link
               when SHIFT_PREV_WAIT =>
                  char_rdy <= '0';
                  if PREV_DONE = '1' then
                     char_state <= SHIFT_D_READ;
                  elsif PREV_RDY = '1' then
                     char_state <= SHIFT_PREV_READ;
                     char_prev_en <= '1';
                  end if;
               when SHIFT_PREV_READ =>
                  char_d <= PREV_DIN;
                  char_rdy <= '1';
                  char_state <= SHIFT_PREV_WRITE;
               when SHIFT_PREV_WRITE =>
                  if ENABLE = '1' then
                     char_rdy <= '0';
                     char_state <= SHIFT_PREV_WAIT;
                  end if;

               -- shift out local data
               when SHIFT_D_READ =>
                  char_d <= char_dout;
                  char_count <= char_count + 1;
                  char_rdy <= '1';
                  char_state <= SHIFT_D_WRITE;
               when SHIFT_D_WRITE =>
                  if ENABLE = '1' then
                     char_rdy <= '0';
                     if char_count = To_Unsigned (INPUT_BITS / 8, 16) then
                        char_done <= '1';
                        char_state <= IDLE;
                     else
                        char_state <= SHIFT_D_READ;
                     end if;
                  end if;

             end case;
         end if;
      end if;
   end process char_fsm;
   char_addr (10 downto 0) <= std_logic_vector (char_count (10 downto 0));

   PREV_START <= START when rising_edge (CLK);
   PREV_ENABLE <= char_prev_en;
   RDY <= char_rdy;
   DONE <= char_done;
   DOUT <= char_d;


   -----------------------------------------------------------------------------
   -- SHIFT FROM INPUT FIBER
   -----------------------------------------------------------------------------

   -- block ram buffer
   buffer_ram: bram_16384_1_8
      port map (
         clka   => CLK,
         dina   => in_din,
         addra  => in_addr,
         wea    => in_we,
         douta  => open,
         clkb   => CLK,
         dinb   => X"00",
         addrb  => char_addr,
         web    => B"0",
         doutb  => char_dout
      );
   in_din(0) <= IN_D;
   in_we(0) <= IN_CE;
   in_addr (13 downto 3) <=     std_logic_vector (in_count (13 downto 3));
   in_addr ( 2 downto 0) <= not std_logic_vector (in_count ( 2 downto 0));

   -- input CRC generator
   input_crc: ucrc_ser
      generic map (
         POLYNOMIAL => CRC32_POLYNOMIAL,
         INIT_VALUE => CRC32_INIT_VALUE,
         SYNC_RESET => 0)
      port map (
         clk_i   => CLK,
         rst_i   => IN_START,
         clken_i => IN_CE,
         data_i  => IN_D,
         flush_i => '0',
         match_o => in_crcmatch,
         crc_o   => open
      );

  -- input state machine
   in_fsm: process (CLK) is
   begin
      if rising_edge (CLK) then
         if RST = '1' then
            in_end <= '0';
            in_state <= SHIFT_DONE;
            in_count <= To_Unsigned (0, 16);
            in_timeout <= (others => '1');
            in_to <= '0';
         elsif IN_START = '1' then
            in_state <= SHIFT_D;
            in_count <= To_Unsigned (0, 16);
            in_to <= '0';
         else
            in_to <= '0';
            case in_state is
               when SHIFT_D =>
                  if IN_CE = '1' then
                     in_count <= in_count + 1;
                     if in_count = To_Unsigned (INPUT_BITS - 1, 16) then
                        in_count <= To_Unsigned (0, 16);
                        in_state <= SHIFT_WAIT;
                    end if;
                  end if;
               when SHIFT_WAIT =>
                  in_count <= in_count + 1;
                  if in_count = To_Unsigned (15, 16) then
                     in_end <= '1';
                     in_timeout <= (others => '1');
                     in_state <= SHIFT_DONE;
                  end if;
               when SHIFT_DONE =>
                  in_end <= '0';
                  in_timeout <= in_timeout - 1;
                  if in_timeout = To_Unsigned (0, 28) then
                     in_to <= '1'; -- reenable 1PPS so we don't get stuck!!!
                  end if;
            end case;
         end if;
      end if;
   end process in_fsm;

   IN_DONE <= in_end or in_to;
   IN_CRCERR <= in_end and not in_crcmatch;
   IN_CRCOK  <= in_end and in_crcmatch;


   -----------------------------------------------------------------------------
   -- SHIFT TO OUTPUT FIBER
   -----------------------------------------------------------------------------

   -- output CRC generator
   output_crc: ucrc_ser
      generic map (
         POLYNOMIAL => CRC32_POLYNOMIAL,
         INIT_VALUE => CRC32_INIT_VALUE,
         SYNC_RESET => 0)
      port map (
         clk_i   => CLK,
         rst_i   => OUT_START,
         clken_i => OUT_CE,
         data_i  => out_data,
         flush_i => out_crc_flush,
         match_o => open,
         crc_o   => out_crc
      );

   -- output bits
   out_data <= out_reg(OUTPUT_BITS-1);
   OUT_D <= out_data when (out_state = SHIFT_D) else out_crc(31);

   -- output state machine
   out_fsm: process (CLK) is
   begin
      if rising_edge (CLK) then
         if RST = '1' then
            out_state <= SHIFT_DONE;
         elsif OUT_START = '1' then
            out_reg(OUTPUT_BITS- 1 downto OUTPUT_BITS-32)  <= GPS;
            out_reg(OUTPUT_BITS-33 downto OUTPUT_BITS-64)  <= ADDR;
            out_reg(OUTPUT_BITS-65 downto OUTPUT_BITS-96)  <= A;
            out_reg(OUTPUT_BITS-97 downto OUTPUT_BITS-128) <= B;
            out_state <= SHIFT_D;
            out_crc_flush <= '0';
            out_count <= To_Unsigned (OUTPUT_BITS - 1, 16);
         else
            case out_state is
               when SHIFT_D =>
                  if OUT_CE = '1' then
                     out_reg <= out_reg (126 downto 0) & '0';
                     out_count <= out_count - 1;
                     if out_count = To_Unsigned (0, 16) then
                        out_state <= SHIFT_CRC;
                        out_crc_flush <= '1';
                        out_count <= To_Unsigned (31, 16);
                     end if;
                  end if;
               when SHIFT_CRC =>
                  if OUT_CE = '1' then
                     out_count <= out_count - 1;
                     if out_count = To_Unsigned (0, 16) then
                        out_state <= SHIFT_DONE;
                     end if;
                  end if;
               when SHIFT_DONE =>
                  -- nothing
                  out_crc_flush <= '0';
            end case;
         end if;
      end if;
   end process out_fsm;

   OUT_DONE <= '1' when (out_state = SHIFT_DONE)
                  else '0';

end architecture Structure;


